/*
 * vector.h - This file contains declarations of vector functions
 *
 *  $Id: vector.h,v 1.3 1997/11/22 10:16:36 johns Exp $
 */

flt VDot(vector *, vector *);
void VCross(vector *, vector *, vector *);
flt VLength(vector *);
void VNorm(vector *);
void VAdd(vector *, vector *, vector *);
void VSub(vector *, vector *, vector *);
void VAddS(flt, vector *, vector *, vector *);
vector Raypnt(ray *, flt);
void VScale(vector * a, flt s); 

void ColorAddS(color * a, color * b, flt s); 
void ColorAccum(color * a, color * b); 
void ColorScale(color * a, flt s); 

